package fr.inria.dioptase.examples;

import java.io.IOException;
import java.util.Vector;

import fr.inria.dioptase.Dioptase;
import fr.inria.dioptase.atomic.Processor;
import fr.inria.dioptase.atomic.Producer;
import fr.inria.dioptase.driver.discovery.BogusDiscoveryDriverLoader;
import fr.inria.dioptase.driver.network.JSENetworkDriverLoader;
import fr.inria.dioptase.driver.storage.JSEStorageDriverLoader;
import fr.inria.dioptase.driver.thing.BogusThingDriverLoader;
import fr.inria.dioptase.stream.DataType;
import fr.inria.dioptase.stream.Stream;
import fr.inria.dioptase.stream.StreamCallback;
import fr.inria.dioptase.stream.StreamItem;
import fr.inria.dioptase.stream.StreamSchema;
import fr.inria.dioptase.stream.StreamUrl;
import fr.inria.dioptase.thing.Sensor;
import fr.inria.dioptase.utils.Logger;
import fr.inria.dioptase.utils.Logger.Level;


/**
 * @author Benjamin Billet
 *
 */
public class DistributedComputation
{
	public static void main(String[] args)
	throws IOException
	{
		Logger.setLevel(Level.INFO);
		
		Dioptase[] dioptases = new Dioptase[5];
		for(int i = 0; i < dioptases.length; i++)
		{
			dioptases[i] = Dioptase.getInstance(5000 + i);
			dioptases[i].initDrivers(new JSENetworkDriverLoader(), new BogusThingDriverLoader(), new JSEStorageDriverLoader(), new BogusDiscoveryDriverLoader());
			dioptases[i].start();
		}
			
		// get the bogus temperature sensor and instanciate a producer
		Sensor sensor = dioptases[0].getThingProperties().getSensor("temperatureSensor");
		Producer producer = dioptases[0].createSensorProducer("temperatureProducer", sensor, 2000);
		
		// deploy two processors with the SUM and COUNT compiled tasks using the deployment service
		Vector params = new Vector();
		params.add("temperature");
		Utils.deployCompiledTask(dioptases[1], "temperatureSum", "Sum", params, new StreamUrl("127.0.0.1", 5000, "temperatureProducer"));
		Utils.deployCompiledTask(dioptases[2], "temperatureCount", "Count", null, new StreamUrl("127.0.0.1", 5000, "temperatureProducer"));
		Processor temperatureSum = dioptases[1].findProcessor("temperatureSum");
		Processor temperatureCount = dioptases[2].findProcessor("temperatureCount");
		
		// deploy a processor with an interpreted task (read the SUM and COUNT outputs and compute the average temperature value) using the deployment service
		Utils.deployInterpretedTask(dioptases[3], "temperatureAvg", getAVERAGEDescription(temperatureSum.getStream(), temperatureCount.getStream()), new StreamSchema("avg", new DataType("temperature", "celcius", "real")), new StreamUrl("127.0.0.1", 5001, "temperatureSum"), new StreamUrl("127.0.0.1", 5002, "temperatureCount"));
		Processor temperatureAvg = dioptases[3].findProcessor("temperatureAvg");
		
		// deploy a processor with an interpreted task (write the value in the output if the average temperature is greater than 5°) using the deployment service
		Utils.deployInterpretedTask(dioptases[4], "temperatureFilter", getFILTERDescription(temperatureAvg.getStream(), "avg"), new StreamSchema("temperature", new DataType("temperature", "celcius", "real")), new StreamUrl("127.0.0.1", 5003, "temperatureAvg"));
		Processor temperatureFilter = dioptases[4].findProcessor("temperatureFilter");
		
		System.out.println("============ Outputs (press enter to exit):");
		
		// read the output of the producer
		producer.getDispatcher().addCallback("tempCallback", new StreamCallback()
		{				
			@Override
			public void newItemReceived(Stream stream, StreamItem item)
			{
				System.out.println(item.getTimestamp() + " - New temperature measurement : " + item.getTuple().getAttributeValue("temperature") + " degrees " + stream.getSchema().getAttributeType("temperature").getUnit());
			}
		});
		
		/*temperatureSum.getDispatcher().addCallback("tempCallback", new StreamCallback()
		{				
			@Override
			public void newItemReceived(Stream stream, StreamItem item)
			{
				System.out.println(item.getTimestamp() + ": " + stream.getName() + " produces " + item.getTuple());
			}
		});
		
		temperatureCount.getDispatcher().addCallback("tempCallback", new StreamCallback()
		{				
			@Override
			public void newItemReceived(Stream stream, StreamItem item)
			{
				System.out.println(item.getTimestamp() + ": " + stream.getName() + " produces " + item.getTuple());
			}
		});*/
		
		// read the output of the AVERAGE task
		temperatureAvg.getDispatcher().addCallback("tempCallback", new StreamCallback()
		{				
			@Override
			public void newItemReceived(Stream stream, StreamItem item)
			{
				System.out.println(item.getTimestamp() + " - Average temperature : " + item.getTuple().getAttributeValue("avg") + " degrees " + stream.getSchema().getAttributeType("avg").getUnit());
			}
		});
		
		// read the output of the FILTER task
		temperatureFilter.getDispatcher().addCallback("tempCallback", new StreamCallback()
		{				
			@Override
			public void newItemReceived(Stream stream, StreamItem item)
			{
				System.out.println(item.getTimestamp() + " - The average temperature is greater than 5 degrees");
			}
		});
		
		System.in.read();
		for(int i = 0; i < dioptases.length; i++)
		{
			dioptases[i].shutdown();
		}
	}
	
	public static String getAVERAGEDescription(Stream sumInput, Stream countInput)
	{
		return
				"init:" +
				"	(defglobal \"lastsum\" 0)" +
				"	(defglobal \"lastcount\" 0)" +
				"work:" +
				"	(def \"diffsum\" (getNewItems \"" + sumInput.getName() + "\"))" +
				"	(def \"diffcount\" (getNewItems \"" + countInput.getName() + "\"))" +
				"	(if (> (size diffsum) 0)" +
				"		(set \"lastsum\" (itemvalue (sub diffsum 0) \"sum\"))" +
				"	)" +
				"	(if (> (size diffcount) 0)" +
				"		(set \"lastcount\" (itemvalue (sub diffcount 0) \"count\"))" +
				"	)" +
				"	(if (> lastcount 0)" +
				"		(write (item (now) \"avg\" (/ lastsum lastcount)))" +
				"	)" +
				"\r\n";
	}
	
	public static String getFILTERDescription(Stream input, String attribute)
	{
		return
				"work:" +
				"	(def \"diff\" (getNewItems \"" + input.getName() + "\"))" +
				"	(if (> (size diff) 0)" +
				"		(foreach \"i\" diff" +
				"			(if (> (itemvalue i \"" + attribute + "\") 5)" +
				"				(write i)\r\n" +
				"		)\r\n" +
				"	)\r\n";
	}
}
